# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 Justin Santa Barbara
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Constraint-Based Scheduler
"""

from nova import db
from nova import flags
from nova import log as logging
from nova.scheduler import datastore
from nova.scheduler import driver
from nova.scheduler import constraint_lib


LOG = logging.getLogger('nova.scheduler.constraint')
FLAGS = flags.FLAGS


class SchedulerConstraint(constraint_lib.Constraint):
    def request(self):
        return self.solver.context['request']

    def request_context(self):
        return self.solver.context['request_context']

    def scheduler(self):
        return self.solver.context['scheduler']

    def datastore(self):
        return self.scheduler().datastore()

    def topology(self):
        context = self.request_context()
        return self.scheduler().topology(context)

    def selectivity(self):
        # We default to an intermediate value here
        return 0.75

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return self.__str__()


class InstanceConstraintFavorLeastLoaded(SchedulerConstraint):
    def __init__(self):
        super(InstanceConstraintFavorLeastLoaded, self).__init__()
        self.db_results = None

    def _get_database_items(self):
        if not self.db_results:
            self.trace(_("get_instance_hosts_sorted"))
            results = self.datastore().get_instance_hosts_sorted(
                                                    self.request_context())

            for result in results:
                self.cache_item(result)

            self.db_results = results

        return self.db_results

    def score_item(self, candidate):
        score = candidate.utilization_score()
        self.trace_score(candidate, score)
        return score

    def get_candidate_iterator(self):
        """Returns an iterator over all items in best-to-least-good order"""
        requested_vcpus = self.request()['vcpus']

        for host in self._get_database_items():
            if host.used + requested_vcpus <= host.capacity:
                score = self.score_item(host)
                yield constraint_lib.Candidate(host.id, score)
            else:
                self.trace(_("skip %s: capacity") % host)


class InstanceConstraintPlaceNearLocation(SchedulerConstraint):
    def __str__(self):
        return "InstanceConstraintPlaceNearLocation:%s" % self.target_location

    def __init__(self, target_location):
        super(InstanceConstraintPlaceNearLocation, self).__init__()
        self.target_location = target_location

    def _score_service(self, service):
        location = datastore.get_service_location(service)
        distance = self.topology().distance(self.target_location, location)
        score = 1 / float(1 + distance)  # Map to (0, 1]
        LOG.debug("%s scores %s" % (location, score))
        return score

    def score_item(self, candidate):
        score = self._score_service(candidate.model)
        self.trace_score(candidate, score)
        return score

    def get_candidate_iterator(self):
        """Returns an iterator over all items in best-to-least-good order"""
        topo = self.topology()
        for location in topo.find_by_distance_from(self.target_location):
            services = self.datastore().service_get_all_by_topic_location(
                                                    self.request_context(),
                                                    'compute',
                                                    location.key)
            LOG.debug(_("Found services at %(location)s => %(services)s")
                      % locals())
            for service in services:
                score = self._score_service(service)
                LOG.debug("%s scores %s" % (service.host, score))
                yield constraint_lib.Candidate(service.id, score)

    def selectivity(self):
        # TODO(justinsb): We need to systematize this a bit better...
        return 0.1  # We expect to be very selective


class VolumeConstraintFavorLeastGigabytes(SchedulerConstraint):
    def __init__(self):
        super(VolumeConstraintFavorLeastGigabytes, self).__init__()
        self.db_results = None

    def _get_database_items(self):
        if not self.db_results:
            self.trace(_("get_volume_hosts_sorted"))
            results = self.datastore().get_volume_hosts_sorted(
                                                    self.request_context())

            for result in results:
                self.cache_item(result)

            self.db_results = results

        return self.db_results

    def score_item(self, candidate):
        return candidate.utilization_score()

    def get_candidate_iterator(self):
        """Returns an iterator over all items in best-to-least-good order"""
        requested_size = self.request()['size']

        for host in self._get_database_items():
            if host.used + requested_size <= host.capacity:
                score = self.score_item(host)
                yield constraint_lib.Candidate(host.id, score)
            else:
                self.trace(_("skip %s: capacity") % host)


class NetworkConstraintFavorLeastNetworks(SchedulerConstraint):
    def __init__(self):
        super(NetworkConstraintFavorLeastNetworks, self).__init__()
        self.db_results = None

    def _get_database_items(self):
        if not self.db_results:
            self.trace(_("get_network_hosts_sorted"))
            results = self.datastore().get_network_hosts_sorted(
                                                    self.request_context())

            for result in results:
                self.cache_item(result)

            self.db_results = results

        return self.db_results

    def score_item(self, candidate):
        return candidate.utilization_score()

    def get_candidate_iterator(self):
        """Returns an iterator over all items in best-to-least-good order"""
        requested_count = 1

        for host in self._get_database_items():
            if host.used_networks + requested_count <= host.capacity_networks:
                score = self.score_item(host)
                yield constraint_lib.Candidate(host.id, score)
            else:
                self.trace(_("skip %s: capacity") % host)


class ConstraintScheduler(driver.Scheduler):
    """Implements constraint-based scheduler"""
    def __init__(self):
        super(ConstraintScheduler, self).__init__()

    def schedule(self, context, topic, *_args, **_kwargs):
        raise NotImplementedError(_("Must implement a %s scheduler") % topic)

    def schedule_run_instance(self, request_context, instance_id,
                              *_args, **_kwargs):
        """Picks a host that is up and has the fewest running instances."""
        instance_ref = db.instance_get(request_context, instance_id)

        solver = constraint_lib.Solver({'request_context': request_context,
                                        'request': instance_ref,
                                        'scheduler': self})
        solver.add_constraint(InstanceConstraintFavorLeastLoaded())

        # NOTE(justinsb): Once we reach a decision on where this metadata
        #  should be, then use that location.  For now, use 'metadata'
        if instance_ref.get('metadata'):
            for item in instance_ref['metadata']:
                key = item['key']
                value = item['value']

                if key == 'openstack:near':
                    topo = self.topology(request_context)
                    node = topo.get_node_by_key(value, create=True)
                    constraint = InstanceConstraintPlaceNearLocation(node)
                    solver.add_constraint(constraint)

        LOG.debug("Metadata: %s" % instance_ref.get('metadata'))
        LOG.debug("Constraints: %s" % solver.constraints)

        for service in solver.get_solutions_iterator():
            if self.service_is_up(service.model):
                host = service.model['host']
                self.datastore().record_instance_scheduled(request_context,
                                                           instance_id,
                                                           host)
                return host

        raise driver.NoValidHost(_("No suitable hosts found"))

    def schedule_create_volume(self, request_context, volume_id,
                               *_args, **_kwargs):
        """Picks a host that is up and has the fewest volumes."""
        volume_ref = db.volume_get(request_context, volume_id)

        solver = constraint_lib.Solver({'request_context': request_context,
                                        'request': volume_ref,
                                        'scheduler': self})
        solver.add_constraint(VolumeConstraintFavorLeastGigabytes())

        for service in solver.get_solutions_iterator():
            if self.service_is_up(service.model):
                host = service.model['host']
                self.datastore().record_volume_scheduled(request_context,
                                                         volume_id,
                                                         host)
                return host

        raise driver.NoValidHost(_("No suitable hosts found"))

    def schedule_set_network_host(self, request_context, *_args, **_kwargs):
        """Picks a host that is up and has the fewest networks."""

        solver = constraint_lib.Solver({'request_context': request_context,
                                        'request': {},
                                        'scheduler': self})
        solver.add_constraint(NetworkConstraintFavorLeastNetworks())

        for service in solver.get_solutions_iterator():
            if self.service_is_up(service.model):
                host = service.model['host']
                self.datastore().record_network_scheduled(request_context,
                                                          host)
                return host

        raise driver.NoValidHost(_("No suitable hosts found"))
