# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Simple Constraint Solver Library
"""

import heapq

from nova import flags
from nova import log as logging


LOG = logging.getLogger('nova.scheduler.constraint')
FLAGS = flags.FLAGS


class Candidate(object):
    def __init__(self, id, score):
        self.id = id
        self.score = score

    def __str__(self):
        return "%s:%s" % (self.id, self.score)


class Constraint(object):
    def __init__(self):
        self.solver = None

    def __str__(self):
        return "%s" % self.__class__.__name__

    def score_item(self, candidate):
        """Scores the 'goodness' of a candidate.

        For acceptable values, the score must be > 0 and <= 1.
        If score <= 0, then the decision is unacceptable"""
        raise NotImplementedError("Must implement score_item")

    def get_candidate_iterator(self):
        """Returns an iterator over all items in best-to-least-good order"""
        raise NotImplementedError("Must implement get_candidate_iterator")

    def selectivity(self):
        """Returns the selectivity of the constraint in (0,1]

        A smaller value means the constraint is likely to match fewer results.
        If the criteria only matches half the items, use 0.5.
        If it only matches one in 10, use 0.1 etc"""
        raise NotImplementedError("Must implement selectivity")

    def can_beat_score(self, score):
        """Checks if we know that we can match/beat the specified score.

        This is an optimization; a typical implementation will check the items
        in the context and check if at least one of them has a score that is
        >= the passed score. If so, we're not going to be the worst criteria,
        so we don't need to get the list of values. See the implementation of
        the Solver for the details of why!
        """
        for item in self.solver.all_cached_items():
            if self.score_item(item) > score:
                return True
        return False

    def set_solver(self, solver):
        if self.solver:
            raise Exception("Tried to double-set solver")
        self.solver = solver

    def trace(self, message):
        self.solver.trace(message)

    def trace_score(self, candidate, score):
        self.trace("%s scored %s => %s" % (self, candidate, score))

    def cache_item(self, item):
        item_id = item.id
        self.solver.cache_item(item, item_id)


def get_selectivity(constraint):
    return constraint.selectivity()


class ConstraintIteratorState(object):
    """Holds the state of iteration over the candidates for a constraint"""
    def __init__(self, constraint, query_values=True):
        self.constraint = constraint

        self.is_done = False
        self.current = None

        if query_values:
            self.iterator = constraint.get_candidate_iterator()
            self.advance()
        else:
            self.iterator = None

    def advance(self):
        next_item = next(self.iterator, None)
        if next_item:
            self.current = next_item
        else:
            self.current = None
            self.is_done = True

    def __str__(self):
        return "%s %s" % (self.constraint, self.current)


class Solver(object):
    """Finds the minimax solution to the constraints

    Formally, we're looking for max(min(score(candidate, constraint))
    where max is over all candidates
    and min is over all constraints.

    NOTE(justinsb): Anyone know how to get TeX into here? :-)"""

    def __init__(self, context):
        self.cached_items = {}
        self.context = context
        self.constraints = []

    def add_constraint(self, constraint):
        self.constraints.append(constraint)
        constraint.set_solver(self)

    def lookup_item(self, item_id):
        item = self.cached_items.get(item_id)
        if not item:
            # This shouldn't be possible
            raise Exception(_("Item not in cache: %s") % item_id)
            #            if self.lookup_function:
            #            self.trace(_("lookup on %s") % (item_id))
            #            item = self.lookup_function(item_id)
            #            self.cache(item, item_id)
            #            else:
            #                raise Error()
        return item

    def cache_item(self, item, item_id):
        self.cached_items[item_id] = item

    def all_cached_items(self):
        return self.cached_items.values()

    def trace(self, message):
        """Records diagnostic information on scheduling decisions.

        Constraint scheduling is not exactly simple. We may in future want to
        provide (admin) API calls that expose the scheduling reasoning
        """
        LOG.debug(message)

    def get_solutions_iterator(self):
        if not self.constraints:
            raise Exception("No constraints provided")

        queue = []

        # We sort so that the most selective criteria is done first
        # This then lets less selective criteria inspect the candidates that
        #  the more selective candidates have chosen first.
        # This is just a (potential) optimization
        constraints = sorted(self.constraints, key=get_selectivity)

        # Find the criteria which is the lowest scoring (worst)
        # Observe that the worst doesn't actually change... the score must get
        #  lower as we iterate through
        worst = None
        states = []
        for constraint in constraints:
            query_values = True

            if worst and constraint.can_beat_score(worst.current.score):
                # This is a sneaky optimization. If we know we're not going
                #  to be the worst, we don't need to query (see code below)
                # If we'd be querying a REST service or DB, this is important
                self.trace(_("Won't query constraint - not the worst: %s") %
                           (constraint))
                query_values = False

            state = ConstraintIteratorState(constraint,
                                            query_values=query_values)
            states.append(state)

            if query_values:
                if not worst or worst.current.score > state.current.score:
                    worst = state

        self.trace(_("Using lowest-scoring criteria: %s") %
                   (worst.constraint))

        while not worst.is_done:
            self.trace(_("Candidate: %s") % (worst))

            # Loop over other constraints to get the score for the candidate
            # We break ties by choosing the candidate with the sum of scores
            min_score = worst.current.score
            tie_break_score = 0

            item_id = worst.current.id
            item = self.lookup_item(item_id)
            for constraint_state in states:
                score = constraint_state.constraint.score_item(item)
                min_score = min(min_score, score)
                tie_break_score += score

            # Heapify is really nice and will look at tie_break_score
            #  if min_score is the same
            heapq.heappush(queue, (-min_score, -tie_break_score, item))

            # Every future item will have a score <= worst.current.score
            while True:
                (head_score, tie_break_score, head_item) = queue[0]
                head_score = -head_score
                #tie_break_score = -tie_break_score
                if head_score <= worst.current.score:
                    break

                heapq.heappop(queue)
                self.trace(_("Yielding: %s") % (head_item))
                yield head_item

            # Advance the iterator
            worst.advance()
            if not worst.is_done:
                self.trace(_("Advanced: %s") % (worst))

        self.trace(_("Reached end of: %s") % (worst.constraint))

        while queue:
            (head_score, tie_break_score, head_item) = heapq.heappop(queue)
            #head_score = -head_score
            #tie_break_score = -tie_break_score
            self.trace(_("Yielding: %s") % (head_item))
            yield head_item
