# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 Justin Santa Barbara
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
DataStore for the scheduler.

Currently backed by the DB.
"""
import datetime

from nova import db
from nova import flags
from nova import log as logging


LOG = logging.getLogger('nova.scheduler.datastore')
FLAGS = flags.FLAGS


def get_service_location(service):
    #TODO(justinsb): Implement configurable location
    #TODO(justinsb): Move this to service class
    location = None  # service.location

    if not location:
        # Fall back to using the DNS name / IP for "poor man's" topology
        location = service.host
        components = location.split('.')
        components.reverse()
        location = '.'.join(components)
    return location


class SchedulerAbstractHostModel(object):
    def __init__(self, id, model, used, capacity):
        self.id = id
        self.model = model
        self.used = used
        self.capacity = capacity

    def __str__(self):
        return '%s %s/%s' % (self.id, self.used, self.capacity)

    def __repr__(self):
        return "Host:%s" % self

    def utilization_score(self):
        # The best machine is one that is unused
        utilization = float(self.used) / float(self.capacity)
        if utilization > 1:
            utilization = 1
        score = 1 - utilization
        return score


class SchedulerInstanceHostModel(SchedulerAbstractHostModel):
    def __init__(self, id, model, used_cores, capacity_cores):
        super(SchedulerInstanceHostModel, self).__init__(id, model,
                                                         used_cores,
                                                         capacity_cores)


class SchedulerVolumeHostModel(SchedulerAbstractHostModel):
    def __init__(self, id, model, used_gigabytes, capacity_gigabytes):
        super(SchedulerVolumeHostModel, self).__init__(id, model,
                                                       used_gigabytes,
                                                       capacity_gigabytes)


class SchedulerNetworkHostModel(SchedulerAbstractHostModel):
    def __init__(self, id, model, used_networks, capacity_networks):
        super(SchedulerNetworkHostModel, self).__init__(id, model,
                                                        used_networks,
                                                        capacity_networks)


class SchedulerDataStore(object):
    def get_instance_hosts_sorted(self, context):
        results = []
        db_results = db.service_get_all_compute_sorted(context)
        for db_result in db_results:
            (model, used_cores) = db_result
            id = model['id']

            # TODO(justinsb): We need a way to know the true capacity
            capacity_cores = FLAGS.max_cores

            item = SchedulerInstanceHostModel(id,
                                              model,
                                              used_cores,
                                              capacity_cores)
            results.append(item)
        # NOTE(justinsb): This is only sorted by score because max is fixed
        return results

    def service_get_all_by_topic(self, context, topic):
        return db.service_get_all_by_topic(context, topic)

    def service_get_all_by_topic_location(self, context, topic, location):
        #TODO(justinsb): We need a call like this once we've agreed it
        #return db.service_get_all_by_topic_location(context, topic, location)

        services = []
        for service in db.service_get_all_by_topic(context, topic):
            if location == get_service_location(service):
                services.append(service)

        return services

    def get_volume_hosts_sorted(self, context):
        results = []
        db_results = db.service_get_all_volume_sorted(context)
        for db_result in db_results:
            (model, used_gigabytes) = db_result
            id = model['id']

            # TODO(justinsb): We need a way to know the true capacity
            capacity_gigabytes = FLAGS.max_gigabytes

            item = SchedulerVolumeHostModel(id,
                                            model,
                                            used_gigabytes,
                                            capacity_gigabytes)
            results.append(item)
        # NOTE(justinsb): This is only sorted by score because max is fixed
        return results

    def get_network_hosts_sorted(self, context):
        results = []
        db_results = db.service_get_all_network_sorted(context)
        for db_result in db_results:
            (model, used_networks) = db_result
            id = model['id']

            # TODO(justinsb): We need a way to know the true capacity
            capacity_networks = FLAGS.max_networks

            item = SchedulerNetworkHostModel(id,
                                             model,
                                             used_networks,
                                             capacity_networks)
            results.append(item)
        # NOTE(justinsb): This is only sorted by score because max is fixed
        return results

    def record_instance_scheduled(self, context, instance_id, host):
        now = datetime.datetime.utcnow()
        db.instance_update(context,
                           instance_id,
                           {'host': host,
                            'scheduled_at': now})

    def record_volume_scheduled(self, context, volume_id, host):
        now = datetime.datetime.utcnow()
        db.volume_update(context,
                         volume_id,
                         {'host': host,
                          'scheduled_at': now})

    def record_network_scheduled(self, context, host):
        pass

    def service_get_all(self, context):
        return db.service_get_all(context)
