# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2010 Openstack, LLC.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Simple Scheduler
"""

from nova import db
from nova import flags
from nova.scheduler import driver
from nova.scheduler import chance

FLAGS = flags.FLAGS
flags.DEFINE_integer("max_cores", 16,
                     "maximum number of instance cores to allow per host")
flags.DEFINE_integer("max_gigabytes", 10000,
                     "maximum number of volume gigabytes to allow per host")
flags.DEFINE_integer("max_networks", 1000,
                     "maximum number of networks to allow per host")


class SimpleScheduler(chance.ChanceScheduler):
    """Implements Naive Scheduler that tries to find least loaded host."""

    def schedule_run_instance(self, context, instance_id, *_args, **_kwargs):
        """Picks a host that is up and has the fewest running instances."""
        instance_ref = db.instance_get(context, instance_id)
        if (instance_ref['availability_zone']
            and ':' in instance_ref['availability_zone']
            and context.is_admin):
            zone, _x, host = instance_ref['availability_zone'].partition(':')
            service = db.service_get_by_args(context.elevated(), host,
                                             'nova-compute')
            if not self.service_is_up(service):
                raise driver.WillNotSchedule(_("Host %s is not alive") % host)

            self.datastore().record_instance_scheduled(context,
                                                       instance_id,
                                                       host)
            return host
        results = self.datastore().get_instance_hosts_sorted(context)
        for result in results:
            service = result.model
            if result.used + instance_ref['vcpus'] > result.capacity:
                raise driver.NoValidHost(_("All hosts have too many cores"))
            if self.service_is_up(service):
                self.datastore().record_instance_scheduled(context,
                                                           instance_id,
                                                           service['host'])
                return service['host']
        raise driver.NoValidHost(_("No hosts found"))

    def schedule_create_volume(self, context, volume_id, *_args, **_kwargs):
        """Picks a host that is up and has the fewest volumes."""
        volume_ref = db.volume_get(context, volume_id)
        if (volume_ref['availability_zone']
            and ':' in volume_ref['availability_zone']
            and context.is_admin):
            zone, _x, host = volume_ref['availability_zone'].partition(':')
            service = db.service_get_by_args(context.elevated(), host,
                                             'nova-volume')
            if not self.service_is_up(service):
                raise driver.WillNotSchedule(_("Host %s not available") % host)

            self.datastore().record_volume_scheduled(context,
                                                     volume_id,
                                                     host)
            return host
        results = self.datastore().get_volume_hosts_sorted(context)
        for result in results:
            service = result.model
            if result.used + volume_ref['size'] > result.capacity:
                raise driver.NoValidHost(_("All hosts have too many "
                                           "gigabytes"))
            if self.service_is_up(service):
                self.datastore().record_volume_scheduled(context,
                                                         volume_id,
                                                         service['host'])
                return service['host']
        raise driver.NoValidHost(_("No hosts found"))

    def schedule_set_network_host(self, context, *_args, **_kwargs):
        """Picks a host that is up and has the fewest networks."""

        results = self.datastore().get_network_hosts_sorted(context)
        for result in results:
            service = result.model
            if result.used + 1 > result.capacity:
                raise driver.NoValidHost(_("All hosts have too many networks"))
            if self.service_is_up(service):
                self.datastore().record_network_scheduled(context,
                                                          service['host'])
                return service['host']
        raise driver.NoValidHost(_("No hosts found"))
