# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright (c) 2011 Justin Santa Barbara
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Topology mapping to expose e.g. same machine, same rack, same room etc
"""

from nova import flags
from nova import log as logging


LOG = logging.getLogger('nova.scheduler.constraint')
FLAGS = flags.FLAGS


class TopologyNode(object):
    def __str__(self):
        return self.key

    def __repr__(self):
        return "TopologyNode:%s" % self.key

    def __init__(self, key):
        self.key = key


class Topology(object):
    def distance(self, a, b):
        """Returns the distance from TopologyNode a to TopologyNode b"""
        raise NotImplementedError()

    def find_by_distance_from(self, a):
        """Returns all the topology nodes ordered with the closest to a first.

        This may (should?) return an iterator"""
        raise NotImplementedError()


class NamedTopology(Topology):
    """A named toplogy uses DNS-like naming e.g. machine3.rack2.roomb.nyc.us

    We assume that the distance between two items is:
      (D - the length of the suffix path they share)
    where D is the maximum distance.

    Trick: We assume all names are reversed, so
        machine3.rack2.roomb.us => us.roomb.rack2.machine3
        Prefix matching is much more natural, in particular for databases

    This is really eday's concept - all credit to him!
    """
    def __init__(self):
        #TODO(justinsb): As we aim for scale, we should switch to a graph
        self.nodes = []
        self.nodes_by_key = {}
        self.max_distance = 0

    def add(self, node):
        # Pre-compute the components
        node.components = node.key.split('.')
        self.nodes.append(node)
        self.nodes_by_key[node.key] = node
        self.max_distance = max(self.max_distance, len(node.components))

    def distance(self, a, b):
        # This is an inefficient implementation
        if not isinstance(a, TopologyNode):
            a = self.get_node_by_key(a)

        if not isinstance(b, TopologyNode):
            b = self.get_node_by_key(b)

        if not a or not b:
            return self.max_distance + 1

        a_components = a.components
        b_components = b.components

        min_length = min(len(a_components), len(b_components))
        d = 0
        while True:
            if d >= min_length:
                break
            if a_components[d] != b_components[d]:
                break
            d = d + 1

        d = self.max_distance - d
        return d

    def find_by_distance_from(self, a):
        # This is an inefficient implementation
        proximities = [(node, self.distance(a, node))
                       for node in self.nodes]

        proximities.sort(key=lambda p: p[1])

        ret = [p[0] for p in proximities]
        return ret

    def get_node_by_key(self, key, create=False):
        node = self.nodes_by_key.get(key)
        if create and not node:
            node = TopologyNode(key)
            self.add(node)
        return node
