# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2011 Justin Santa Barbara
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Constraint Library
"""

import random

from nova import flags
from nova import log as logging
from nova import test
from nova.scheduler import constraint_lib


LOG = logging.getLogger('nova.tests.constraintlib')

FLAGS = flags.FLAGS
FLAGS.verbose = True


def brute_compute_score(item, constraints):
    min_score = None
    for constraint in constraints:
        score = constraint.score_item(item)
        if min_score and min_score <= score:
            continue
        min_score = score

    return min_score


def brute_force(all_items, constraints):
    max_min_score = None
    max_min_item = None

    for item in all_items:
        min_score = brute_compute_score(item, constraints)

        if max_min_score and max_min_score >= min_score:
            continue

        max_min_score = min_score
        max_min_item = item

    return (max_min_item, max_min_score)


class ParameterBoundConstraint(constraint_lib.Constraint):
    def __init__(self, all_items, key, selectivity_value):
        super(ParameterBoundConstraint, self).__init__()
        self.all_items = all_items
        self.key = key
        self.selectivity_value = selectivity_value

    def to_s(self):
        return "ParameterBoundConstraint:%s" % self.key

    def score_item(self, candidate):
        return candidate[self.key]

    def get_candidate_iterator(self):
        """Returns an iterator over all items in best-to-least-good order"""
        all_items_ordered = sorted(self.all_items,
                                   key=self.score_item,
                                   reverse=True)
        for item in all_items_ordered:
            score = self.score_item(item)
            item_id = item['id']
            yield constraint_lib.Candidate(item_id, score)

    def selectivity(self):
        return self.selectivity_value


class ConstraintLibTestCase(test.TestCase):
    """Test case for constraint library"""

    def test_five_five__five(self):
        rnd = random.Random()
        rnd.seed(555)

        for _rep in range(5):
            self._do_random_test(rnd, 5, 5)

    def test_big(self):
        rnd = random.Random()
        rnd.seed(1234)

        item_count = 1000
        constraints_count = 100

        self._do_random_test(rnd, item_count, constraints_count)

    def _test_random_100(self):
        """This is hidden because it's slow, but it's a nice torture test"""
        rnd = random.Random()
        rnd.seed(1234)

        for _rep in range(100):
            item_count = rnd.randint(100, 1000)
            constraints_count = rnd.randint(5, 100)

            self._do_random_test(rnd, item_count, constraints_count)

    def _do_random_test(self, rnd, item_count, constraints_count):
        all_items = []

        context = {}
        solver = constraint_lib.Solver(context)

        for j in range(item_count):
            item = {}
            for i in range(constraints_count):
                key = 'c%s' % i
                item[key] = rnd.random()
            item['id'] = j
            all_items.append(item)
            solver.cache_item(item, item['id'])

        constraints = []
        for i in range(constraints_count):
            key = 'c%s' % i
            selectivity_value = rnd.random()
            constraint = ParameterBoundConstraint(all_items,
                                                  key,
                                                  selectivity_value)
            solver.add_constraint(constraint)
            constraints.append(constraint)

        solved_item = None

        previous_item = None
        for item in solver.get_solutions_iterator():
            if not solved_item:
                solved_item = item
            if previous_item:
                previous_score = brute_compute_score(previous_item,
                                                     constraints)
                item_score = brute_compute_score(item, constraints)

                if previous_score < item_score:
                    LOG.warn("PREVIOUS: %s %s" % (previous_item,
                                                  previous_score))
                    LOG.warn("THIS: %s %s" % (item,
                                                  item_score))
                    self.assertFalse("Items not returned in order")

            previous_item = item

        (brute_item, brute_score) = brute_force(all_items, constraints)

        self.assertEquals(brute_score,
                          brute_compute_score(brute_item, constraints))
        self.assertEquals(brute_score,
                          brute_compute_score(solved_item, constraints))
        self.assertEquals(brute_item, solved_item)
